package io.github.serpro69.kfaker.provider.misc

import io.github.serpro69.kfaker.FakerConfig
import io.github.serpro69.kfaker.fakerConfig
import io.kotest.assertions.assertSoftly
import io.kotest.assertions.throwables.shouldNotThrow
import io.kotest.assertions.throwables.shouldThrow
import io.kotest.core.spec.style.DescribeSpec
import io.kotest.matchers.collections.shouldHaveSize
import io.kotest.matchers.maps.shouldHaveSize
import io.kotest.matchers.shouldBe
import io.kotest.matchers.shouldNotBe
import io.kotest.matchers.string.shouldHaveLength
import io.kotest.matchers.types.instanceOf
import io.kotest.matchers.types.shouldNotBeSameInstanceAs
import java.util.*
import kotlin.reflect.full.declaredMemberProperties

@Suppress("unused")
class RandomClassProviderTest : DescribeSpec({
    assertSoftly = true

    val config = fakerConfig { random = Random() }
    val randomProvider = RandomClassProvider(config)

    describe("a TestClass with an empty constructor") {
        class TestClass

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with non-empty constructor") {
        class Foo
        class TestClass(val foo: Foo)

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with non-empty constructor with object type parameter") {
        class TestClass(val testObject: TestObject)

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with non-empty constructor with primitive type parameters") {
        class TestClass(
            val double: Double,
            val float: Float,
            val long: Long,
            val int: Int,
            val short: Short,
//            val byte: Byte, // TODO: 11.06.19
            val string: String,
            val char: Char,
            val boolean: Boolean
//            val array: Array<String> // TODO: 12.06.19
        )

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with non-empty constructor with primitive type and custom-type parameters") {
        class Foo(val string: String)

        class TestClass(
            val foo: Foo,
            val double: Double,
            val float: Float,
            val long: Long,
            val int: Int,
            val short: Short,
            val byte: Byte,
            val string: String,
            val char: Char,
            val boolean: Boolean
        )

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with non-empty constructor with nullable type parameters") {
        class Foo(val string: String?)
        class TestClass(val foo: Foo?)

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with private constructor") {
        class TestClass private constructor()

        context("creating a random instance of the class") {
            it("exception is thrown") {
                val exception = shouldThrow<NoSuchElementException> {
                    randomProvider.randomClassInstance<TestClass>()
                }

                exception.message shouldBe "No suitable constructor found for ${TestClass::class}"
            }
        }
    }

    describe("a TestClass with non-empty constructor with enum type param") {
        class TestClass(val enum: TestEnum)

        context("creating a random instance of the class") {
            val testClass: TestClass = randomProvider.randomClassInstance()

            it("it should be instance of TestClass") {
                testClass shouldBe instanceOf(TestClass::class)
            }
        }
    }

    describe("a TestClass with non-empty constructor with sealed type param") {
        class TestClass(val sealed: TestSealedCls)

        repeat(10) {
            context("creating a random instance of the class run#$it") {
                val testClass: TestClass = randomProvider.randomClassInstance()

                it("it should be instance of TestClass") {
                    testClass shouldBe instanceOf(TestClass::class)
                }
            }
        }
    }

    describe("a TestClass with non-empty constructor for custom type generators") {
        class Foo(val int: Int)
        class TestClass(
            val id: UUID,
            val int: Int,
            val long: Long,
            val foo: Foo
        )

        val givenUuid = UUID.fromString("00000000-0000-0000-0000-000000000000")
        val givenInt = 0

        context("creating a random instance of the class with custom generators") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                typeGenerator<UUID> { givenUuid }
                @Suppress("RemoveExplicitTypeArguments")
                typeGenerator<Int> { givenInt }
            }

            it("it should be a predefined UUID and primitives") {
                assertSoftly {
                    testClass shouldBe instanceOf(TestClass::class)
                    testClass.id shouldBe givenUuid
                    testClass.int shouldBe givenInt
                    testClass.foo.int shouldBe givenInt
                }
            }
        }
    }

    describe("a TestClass with non-empty constructor for custom parameter generators") {
        class TestClass(
            val id: UUID,
            val id2: UUID,
            val nullableId: UUID?,
            val nullableId2: UUID?
        )

        val typeGeneratedId = UUID.fromString("00000000-0000-0000-0000-000000000000")
        val namedParameterGeneratedId = UUID.randomUUID()
        val nullableNamedParameterGeneratedId = UUID.randomUUID()

        context("creating a random instance of the class with custom parameter and type generators") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                namedParameterGenerator("id") { namedParameterGeneratedId }
                namedParameterGenerator("nullableId") { nullableNamedParameterGeneratedId }
                namedParameterGenerator("nullableId2") { null }
                typeGenerator<UUID> { typeGeneratedId }
            }

            it("should use predefined UUIDs with parameter generators taking precedence over type generators") {
                assertSoftly {
                    testClass shouldBe instanceOf(TestClass::class)
                    testClass.id shouldBe namedParameterGeneratedId
                    testClass.id2 shouldBe typeGeneratedId
                    testClass.nullableId shouldBe nullableNamedParameterGeneratedId
                    testClass.nullableId2 shouldBe null
                }
            }
        }
    }

    describe("a TestClass with non-empty constructor for nullable custom type generators") {
        class Foo(val int: Int?)
        class TestClass(
            val id: UUID,
            val int: Int?,
            val nullableLong: Long?,
            val notNullLong: Long,
            val foo: Foo
        )

        val givenUuid = UUID.fromString("00000000-0000-0000-0000-000000000000")
        val givenInt = 1
        val givenNullableLong: Long? = null
        val givenLong = 1L

        context("creating a random instance of the class with nullable custom generators") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                @Suppress("RemoveExplicitTypeArguments")
                typeGenerator<UUID> { givenUuid }
                @Suppress("RemoveExplicitTypeArguments")
                typeGenerator<Int> { givenInt } // use the not-null type generator and verify it is used for Int?
                nullableTypeGenerator<Long?> { givenNullableLong }
                @Suppress("RemoveExplicitTypeArguments")
                typeGenerator<Long> { givenLong }
            }

            it("it should be a predefined UUID and primitives with nulls") {
                assertSoftly {
                    testClass shouldBe instanceOf(TestClass::class)
                    testClass.id shouldBe givenUuid
                    testClass.int shouldBe givenInt
                    testClass.nullableLong shouldBe givenNullableLong
                    testClass.notNullLong shouldBe givenLong
                    testClass.foo.int shouldBe givenInt
                }
            }
        }
    }

    describe("a TestClass with 3 non-default constructors") {

        class Foo
        class Bar(val int: Int)
        class Baz(val foo: Foo, val string: String)

        class TestClass {
            var foo: Foo? = null
                private set
            var bar: Bar? = null
                private set
            var baz: Baz? = null
                private set

            constructor()

            constructor(foo: Foo) {
                this.foo = foo
            }

            constructor(foo: Foo, bar: Bar) : this(foo) {
                this.bar = bar
            }

            constructor(foo: Foo, bar: Bar, baz: Baz) : this(foo, bar) {
                this.baz = baz
            }
        }

        context("creating a random instance of the class with constructor args size") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                constructorParamSize = 3
            }

            it("constructor with specified args size is used") {
                assertSoftly {
                    testClass.foo shouldNotBe null
                    testClass.bar shouldNotBe null
                    testClass.baz shouldNotBe null
                }
            }
        }

        context("creating a random instance of the class with constructor filter strategy") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                constructorFilterStrategy = ConstructorFilterStrategy.MAX_NUM_OF_ARGS
            }

            it("constructor with specified filter is used") {
                assertSoftly {
                    testClass.foo shouldNotBe null
                    testClass.bar shouldNotBe null
                    testClass.baz shouldNotBe null
                }
            }
        }

        context("creating a random instance of the class with invalid number of args and a fallback strategy") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                constructorParamSize = 6
                fallbackStrategy = FallbackStrategy.USE_MAX_NUM_OF_ARGS
            }

            it("constructor with specified fallback is used") {
                assertSoftly {
                    testClass.foo shouldNotBe null
                    testClass.bar shouldNotBe null
                    testClass.baz shouldNotBe null
                }
            }
        }

        context("creating a random instance of the class with invalid number of args and a filter strategy") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                constructorParamSize = 6
                constructorFilterStrategy = ConstructorFilterStrategy.MAX_NUM_OF_ARGS
            }

            it("constructor with specified fallback is used") {
                assertSoftly {
                    testClass.foo shouldNotBe null
                    testClass.bar shouldNotBe null
                    testClass.baz shouldNotBe null
                }
            }
        }

        context("precedence of using constructor param size over a filter strategy") {
            val testClass: TestClass = randomProvider.randomClassInstance {
                constructorParamSize = 2
                constructorFilterStrategy = ConstructorFilterStrategy.MAX_NUM_OF_ARGS
            }

            it("constructor with specified fallback is used") {
                assertSoftly {
                    testClass.foo shouldNotBe null
                    testClass.bar shouldNotBe null
                    testClass.baz shouldBe null
                }
            }
        }
    }

    describe("a TestClass with non-empty constructor with Collection-type parameters") {
        class Foo
        class Bar(val int: Int)
        class Baz(val foo: Foo, val string: String)
        class TestClass(
            val list: List<Foo>,
            val set: Set<Bar>,
            val map: Map<String, Baz>
        )

        it("should generate Collections with default size 1") {
            val testClass = randomProvider.randomClassInstance<TestClass>()
            assertSoftly {
                testClass.list shouldHaveSize 1
                testClass.set shouldHaveSize 1
                testClass.map shouldHaveSize 1
            }
        }

        it("should generate Collections with pre-configured size") {
            val testClass = randomProvider.randomClassInstance<TestClass> {
                collectionsSize = 10
            }
            assertSoftly {
                testClass.list shouldHaveSize 10
                testClass.set shouldHaveSize 10
                testClass.map shouldHaveSize 10
            }
        }
        it("should generate Collections with pre-configured type generation") {
            val testClass = randomProvider.randomClassInstance<TestClass> {
                typeGenerator<List<Foo>> { listOf() }
            }
            assertSoftly {
                testClass.list shouldHaveSize 0
                testClass.set shouldHaveSize 1
                testClass.map shouldHaveSize 1
            }
        }
    }

    describe("a primitive type") {
        it("should not throw an exception when no suitable constructor exists") {
            assertSoftly {
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Double>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Float>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Long>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Int>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Short>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Byte>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Boolean>() }
                shouldNotThrow<NoSuchElementException> { randomProvider.randomClassInstance<Char>() }
            }
        }

        it("should generate a random string") {
            randomProvider.randomClassInstance<String>() shouldHaveLength 24
        }
    }

    describe("RandomClassProvider configuration") {
        class Foo(val int: Int)
        class Bar(val foo: Foo)
        class Baz(val foo: Foo, val bar: Bar, val test: Int, val nullable: String?)

        val cfg: () -> FakerConfig = {
            fakerConfig {
                randomClassInstance {
                    typeGenerator { Foo(0) }
                    nullableTypeGenerator { "nn" }
                    namedParameterGenerator("test") { 1 }
                }
            }
        }

        it("should configure all random instances") {
            with(RandomClassProvider(config)) {
                configure {
                    typeGenerator { Foo(42) }
                    nullableTypeGenerator { "nn" }
                    namedParameterGenerator("test") { 32167 }
                }
                randomClassInstance<Bar>().foo.int shouldBe 42
                randomClassInstance<Baz>().foo.int shouldBe 42
                randomClassInstance<Baz>().bar.foo.int shouldBe 42
                randomClassInstance<Baz>().test shouldBe 32167
                randomClassInstance<Baz>().nullable shouldBe "nn"
            }
        }

        context("precedence") {
            it("should configure random providers from fakerConfig") {
                with(RandomClassProvider(cfg())) {
                    randomClassInstance<Bar>().foo.int shouldBe 0
                    randomClassInstance<Baz>().foo.int shouldBe 0
                    randomClassInstance<Baz>().bar.foo.int shouldBe 0
                    randomClassInstance<Baz>().test shouldBe 1
                    randomClassInstance<Baz>().nullable shouldBe "nn"
                }
            }
            it("should have defaults if not configured on fakerConfig level") {
                with(RandomClassProvider(fakerConfig { })) {
                    randomClassInstance<Bar>().foo.int shouldNotBe 0
                    randomClassInstance<Baz>().foo.int shouldNotBe 0
                    randomClassInstance<Baz>().bar.foo.int shouldNotBe 0
                    randomClassInstance<Baz>().test shouldNotBe 1
                    randomClassInstance<Baz>().nullable shouldNotBe "nn"
                }
            }
            it("should override configuration from RandomProvider level") {
                with(RandomClassProvider(cfg())) {
                    configure {
                        typeGenerator { Foo(42) }
                        nullableTypeGenerator { "just a string" }
                        namedParameterGenerator("test") { 32167 }
                    }
                    randomClassInstance<Bar>().foo.int shouldBe 42
                    randomClassInstance<Baz>().foo.int shouldBe 42
                    randomClassInstance<Baz>().bar.foo.int shouldBe 42
                    randomClassInstance<Baz>().test shouldBe 32167
                    randomClassInstance<Baz>().nullable shouldBe "just a string"
                }
            }
            it("should re-configure random providers") {
                with(RandomClassProvider(cfg())) {
                    configure {
                        typeGenerator { Foo(42) }
                        nullableTypeGenerator { "just a string" }
                        namedParameterGenerator("test") { 32167 }
                    }
                    randomClassInstance<Bar>().foo.int shouldBe 42
                    randomClassInstance<Baz>().foo.int shouldBe 42
                    randomClassInstance<Baz>().bar.foo.int shouldBe 42
                    randomClassInstance<Baz>().test shouldBe 32167
                    randomClassInstance<Baz>().nullable shouldBe "just a string"

                    configure {
                        typeGenerator { Foo(12177) }
                        nullableTypeGenerator<String> { null }
                    }
                    randomClassInstance<Bar>().foo.int shouldBe 12177
                    randomClassInstance<Baz>().foo.int shouldBe 12177
                    randomClassInstance<Baz>().bar.foo.int shouldBe 12177
                    randomClassInstance<Baz>().test shouldBe 32167
                    randomClassInstance<Baz>().nullable shouldBe null
                }
            }
            it("should use own configuration on function level") {
                with(RandomClassProvider(cfg())) {
                    configure {
                        typeGenerator { Foo(12177) }
                        namedParameterGenerator("test") { 32167 }
                        nullableTypeGenerator<String> { null }
                    }
                    randomClassInstance<Baz> { typeGenerator { Foo(36) } }.also {
                        it.foo.int shouldBe 36
                        it.bar.foo.int shouldBe 36
                        it.test shouldNotBe 1
                        it.test shouldNotBe 32167
                        it.nullable shouldNotBe "nn"
                        it.nullable shouldNotBe null
                    }
                }
            }
        }

        context("resetting to defaults") {
            it("should reset configuration to defaults on RandomProvider level") {
                with(RandomClassProvider(cfg())) {
                    configure { reset() }
                    randomClassInstance<Bar>().foo.int shouldNotBe 0
                    randomClassInstance<Baz>().foo.int shouldNotBe 0
                    randomClassInstance<Baz>().bar.foo.int shouldNotBe 0
                    randomClassInstance<Baz>().test shouldNotBe 1
                    randomClassInstance<Baz>().nullable shouldNotBe "nn"
                }
            }
            it("should re-configure random providers") {
                with(RandomClassProvider(cfg())) {
                    configure {
                        reset()
                        typeGenerator { Foo(12177) }
                        nullableTypeGenerator<String> { null }
                    }
                    randomClassInstance<Bar>().foo.int shouldBe 12177
                    randomClassInstance<Baz>().foo.int shouldBe 12177
                    randomClassInstance<Baz>().bar.foo.int shouldBe 12177
                    randomClassInstance<Baz>().test shouldNotBe 1
                    randomClassInstance<Baz>().nullable shouldBe null
                }
            }
/*            it("should reset configuration to defaults on fakerConfig level") {
                val c = cfg()
                with(RandomClassProvider(c)) {
                    configure {
                        typeGenerator { Foo(12177) }
                        nullableTypeGenerator<String> { null }
                    }
                    c.randomProviderConfig?.reset()
                    randomClassInstance<Bar>().foo.int shouldNotBe 12177
                    randomClassInstance<Baz>().foo.int shouldNotBe 12177
                    randomClassInstance<Baz>().bar.foo.int shouldNotBe 12177
                    randomClassInstance<Baz>().test shouldNotBe 1
                    randomClassInstance<Baz>().nullable shouldNotBe null
                }
            }*/
        }
    }

    describe("new instance of RandomClassProvider") {
        val cfg: () -> FakerConfig = {
            fakerConfig {
                randomClassInstance {
                    collectionsSize = random.nextInt()
                    constructorParamSize = random.nextInt()
                    constructorFilterStrategy = ConstructorFilterStrategy.MAX_NUM_OF_ARGS
                    fallbackStrategy = FallbackStrategy.FAIL_IF_NOT_FOUND
                }
            }
        }
        val configProperties: (RandomProviderConfig) -> List<Any?> = {
            it::class.declaredMemberProperties.map { p -> p.call(it) }
        }
        it("should create a new instance") {
            val rcp = RandomClassProvider(cfg())
            rcp shouldNotBeSameInstanceAs rcp.new()
        }
        it("should have same configuration as defined on fakerConfig level") {
            val c = cfg()
            with(RandomClassProvider(c)) {
                configure {
                    reset()
                    collectionsSize = 100
                    constructorParamSize = 10
                    constructorFilterStrategy = ConstructorFilterStrategy.MIN_NUM_OF_ARGS
                    fallbackStrategy = FallbackStrategy.USE_MIN_NUM_OF_ARGS
                    typeGenerator { "42" }
                    typeGenerator { 42 }
                    typeGenerator { 42.0 }
                }
                val new = new()
                configProperties(new.config) shouldBe configProperties(c.randomProviderConfig!!)
                new.config shouldNotBeSameInstanceAs c.randomProviderConfig
                new.config shouldNotBeSameInstanceAs this@with.config
            }
        }
        it("should have default configuration if none was defined on fakerConfig level") {
            val c = fakerConfig { }
            with(RandomClassProvider(c)) {
                configure {
                    reset()
                    collectionsSize = 100
                    constructorParamSize = 10
                    constructorFilterStrategy = ConstructorFilterStrategy.MIN_NUM_OF_ARGS
                    fallbackStrategy = FallbackStrategy.USE_MIN_NUM_OF_ARGS
                    typeGenerator { "42" }
                    typeGenerator { 42 }
                    typeGenerator { 42.0 }
                }
                val new = new()
                configProperties(new.config) shouldBe configProperties(RandomProviderConfig())
                new.config shouldNotBeSameInstanceAs this@with.config
                new.config shouldNotBeSameInstanceAs c.randomProviderConfig
            }
        }
    }

    context("copy of RandomClassProvider") {
        val cfg: () -> FakerConfig = {
            fakerConfig {
                randomClassInstance {
                    collectionsSize = random.nextInt()
                    constructorParamSize = random.nextInt()
                    constructorFilterStrategy = ConstructorFilterStrategy.MAX_NUM_OF_ARGS
                    fallbackStrategy = FallbackStrategy.FAIL_IF_NOT_FOUND
                }
            }
        }
        val configProperties: (RandomProviderConfig) -> List<Any?> = {
            it::class.declaredMemberProperties.map { p -> p.call(it) }
        }
        it("should have same configuration as defined on RandomClassProvider level") {
            val c = cfg()
            with(RandomClassProvider(c)) {
                configure {
                    collectionsSize = 100
                    constructorParamSize = 10
                    constructorFilterStrategy = ConstructorFilterStrategy.MIN_NUM_OF_ARGS
                    fallbackStrategy = FallbackStrategy.USE_MIN_NUM_OF_ARGS
                    typeGenerator { "42" }
                    typeGenerator { 42 }
                    typeGenerator { 42.0 }
                }
                val copy = copy()
                configProperties(copy.config) shouldBe configProperties(this@with.config)
                copy.config shouldNotBeSameInstanceAs this@with.config
                copy.config shouldNotBeSameInstanceAs c.randomProviderConfig
            }
        }
        it("should have same configuration as defined on fakerConfig level if not defined on RandomClassProvider level") {
            val c = cfg()
            with(RandomClassProvider(c)) {
                val copy = copy()
                configProperties(copy.config) shouldBe configProperties(c.randomProviderConfig!!)
                copy.config shouldNotBeSameInstanceAs c.randomProviderConfig
                copy.config shouldNotBeSameInstanceAs this@with.config
            }
        }
    }
})

@Suppress("unused")
enum class TestEnum {
    KOTLIN,
    JAVA,
    GO
}

@Suppress("CanSealedSubClassBeObject", "unused")
sealed class TestSealedCls {
    object Kotlin : TestSealedCls()
    class Java : TestSealedCls()
}

@Suppress("unused")
class Go(val name: String) : TestSealedCls()

object TestObject
